@@ -1,34 +1,23 @@
 /*
- *   Copyright 2019 Marco Martin <mart@kde.org>
+ *  SPDX-FileCopyrightText: 2019 Marco Martin <mart@kde.org>
  *
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU Library General Public License as
- *   published by the Free Software Foundation; either version 2, or
- *   (at your option) any later version.
- *
- *   This program is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details
- *
- *   You should have received a copy of the GNU Library General Public
- *   License along with this program; if not, write to the
- *   Free Software Foundation, Inc.,
- *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ *  SPDX-License-Identifier: LGPL-2.0-or-later
  */
 
 #include "columnview.h"
 #include "columnview_p.h"
 
+#include "loggingcategory.h"
 #include <QAbstractItemModel>
+#include <QDebug>
 #include <QGuiApplication>
-#include <QStyleHints>
+#include <QPropertyAnimation>
 #include <QQmlComponent>
 #include <QQmlContext>
 #include <QQmlEngine>
-#include <QDebug>
-#include <QPropertyAnimation>
+#include <QStyleHints>
 
+#include "units.h"
 
 QHash<QObject *, ColumnViewAttached *> ColumnView::m_attachedObjects = QHash<QObject *, ColumnViewAttached *>();
 
@@ -36,15 +25,16 @@ class QmlComponentsPoolSingleton
 {
 public:
     QmlComponentsPoolSingleton()
-    {}
+    {
+    }
     static QmlComponentsPool *instance(QQmlEngine *engine);
+
 private:
-    QHash<QQmlEngine*, QmlComponentsPool*> m_instances;
+    QHash<QQmlEngine *, QmlComponentsPool *> m_instances;
 };
 
 Q_GLOBAL_STATIC(QmlComponentsPoolSingleton, privateQmlComponentsPoolSelf)
 
-
 QmlComponentsPool *QmlComponentsPoolSingleton::instance(QQmlEngine *engine)
 {
     Q_ASSERT(engine);
@@ -56,7 +46,7 @@ QmlComponentsPool *QmlComponentsPoolSingleton::instance(QQmlEngine *engine)
 
     componentPool = new QmlComponentsPool(engine);
 
-    QObject::connect(componentPool, &QObject::destroyed, [engine]() {
+    QObject::connect(componentPool, &QObject::destroyed, engine, [engine]() {
         if (privateQmlComponentsPoolSelf) {
             privateQmlComponentsPoolSelf->m_instances.remove(engine);
         }
@@ -70,27 +60,36 @@ QmlComponentsPool::QmlComponentsPool(QQmlEngine *engine)
 {
     QQmlComponent *component = new QQmlComponent(engine, this);
 
-    component->setData(QByteArrayLiteral("import QtQuick 2.7\n"
-        "import org.kde.kirigami 2.7 as Kirigami\n"
-        "QtObject {\n"
-            "id: root\n"
-            "readonly property Kirigami.Units units: Kirigami.Units\n"
-            "readonly property Component separator: Kirigami.Separator {"
-                "property Item column\n"
-                "visible: column.Kirigami.ColumnView.view.contentX < column.x;"
-                "anchors.top: column.top;"
-                "anchors.bottom: column.bottom;"
-            "}\n"
-            "readonly property Component rightSeparator: Kirigami.Separator {"
-                "property Item column\n"
-                "anchors.top: column.top;"
-                "anchors.right: column.right;"
-                "anchors.bottom: column.bottom;"
-            "}"
-        "}"), QUrl());
+    /* clang-format off */
+    component->setData(QByteArrayLiteral(R"(
+import QtQuick 2.7
+import org.kde.kirigami 2.7 as Kirigami
+
+QtObject {
+    id: root
+    readonly property Kirigami.Units units: Kirigami.Units
+
+    readonly property Component separator: Kirigami.Separator {
+        property Item column
+
+        visible: column.Kirigami.ColumnView.view && column.Kirigami.ColumnView.view.contentX < column.x
+        anchors.top: column.top
+        anchors.bottom: column.bottom
+    }
+
+    readonly property Component rightSeparator: Kirigami.Separator {
+        property Item column
+
+        anchors.top: column.top
+        anchors.right: column.right
+        anchors.bottom: column.bottom
+    }
+}
+)"), QUrl(QStringLiteral("columnview.cpp")));
+    /* clang-format on */
 
     m_instance = component->create();
-    //qWarning()<<component->errors();
+    // qCWarning(KirigamiLog)<<component->errors();
     Q_ASSERT(m_instance);
 
     m_separatorComponent = m_instance->property("separator").value<QQmlComponent *>();
@@ -99,25 +98,27 @@ QmlComponentsPool::QmlComponentsPool(QQmlEngine *engine)
     m_rightSeparatorComponent = m_instance->property("rightSeparator").value<QQmlComponent *>();
     Q_ASSERT(m_rightSeparatorComponent);
 
-    m_units = m_instance->property("units").value<QObject *>();
+    m_units = engine->singletonInstance<Kirigami::Units *>(qmlTypeId("org.kde.kirigami", 2, 0, "Units"));
     Q_ASSERT(m_units);
 
-    connect(m_units, SIGNAL(gridUnitChanged()), this, SIGNAL(gridUnitChanged()));
-    connect(m_units, SIGNAL(longDurationChanged()), this, SIGNAL(longDurationChanged()));
+    connect(m_units, &Kirigami::Units::gridUnitChanged, this, &QmlComponentsPool::gridUnitChanged);
+    connect(m_units, &Kirigami::Units::longDurationChanged, this, &QmlComponentsPool::longDurationChanged);
 }
 
 QmlComponentsPool::~QmlComponentsPool()
-{}
-
+{
+}
 
 /////////
 
 ColumnViewAttached::ColumnViewAttached(QObject *parent)
     : QObject(parent)
-{}
+{
+}
 
 ColumnViewAttached::~ColumnViewAttached()
-{}
+{
+}
 
 void ColumnViewAttached::setIndex(int index)
 {
@@ -125,7 +126,7 @@ void ColumnViewAttached::setIndex(int index)
         const bool oldFillWidth = m_fillWidth;
         m_fillWidth = index == m_view->count() - 1;
         if (oldFillWidth != m_fillWidth) {
-            emit fillWidthChanged();
+            Q_EMIT fillWidthChanged();
         }
     }
 
@@ -134,7 +135,7 @@ void ColumnViewAttached::setIndex(int index)
     }
 
     m_index = index;
-    emit indexChanged();
+    Q_EMIT indexChanged();
 }
 
 int ColumnViewAttached::index() const
@@ -154,7 +155,7 @@ void ColumnViewAttached::setFillWidth(bool fill)
     }
 
     m_fillWidth = fill;
-    emit fillWidthChanged();
+    Q_EMIT fillWidthChanged();
 
     if (m_view) {
         m_view->polish();
@@ -183,7 +184,7 @@ void ColumnViewAttached::setReservedSpace(qreal space)
     }
 
     m_reservedSpace = space;
-    emit reservedSpaceChanged();
+    Q_EMIT reservedSpaceChanged();
 
     if (m_view) {
         m_view->polish();
@@ -210,18 +211,18 @@ void ColumnViewAttached::setView(ColumnView *view)
         m_fillWidth = m_index == m_view->count() - 1;
         connect(m_view.data(), &ColumnView::countChanged, this, [this]() {
             m_fillWidth = m_index == m_view->count() - 1;
-            emit fillWidthChanged();
+            Q_EMIT fillWidthChanged();
         });
     }
     if (!m_customReservedSpace && m_view) {
         m_reservedSpace = m_view->columnWidth();
         connect(m_view.data(), &ColumnView::columnWidthChanged, this, [this]() {
             m_reservedSpace = m_view->columnWidth();
-            emit reservedSpaceChanged();
+            Q_EMIT reservedSpaceChanged();
         });
     }
 
-    emit viewChanged();
+    Q_EMIT viewChanged();
 }
 
 QQuickItem *ColumnViewAttached::originalParent() const
@@ -256,7 +257,7 @@ void ColumnViewAttached::setPreventStealing(bool prevent)
     }
 
     m_preventStealing = prevent;
-    emit preventStealingChanged();
+    Q_EMIT preventStealingChanged();
 }
 
 bool ColumnViewAttached::isPinned() const
@@ -272,29 +273,43 @@ void ColumnViewAttached::setPinned(bool pinned)
 
     m_pinned = pinned;
 
-    emit pinnedChanged();
+    Q_EMIT pinnedChanged();
 
     if (m_view) {
         m_view->polish();
     }
 }
 
+bool ColumnViewAttached::inViewport() const
+{
+    return m_inViewport;
+}
 
+void ColumnViewAttached::setInViewport(bool inViewport)
+{
+    if (m_inViewport == inViewport) {
+        return;
+    }
+
+    m_inViewport = inViewport;
+
+    Q_EMIT inViewportChanged();
+}
 
 /////////
 
 ContentItem::ContentItem(ColumnView *parent)
-    : QQuickItem(parent),
-      m_view(parent)
+    : QQuickItem(parent)
+    , m_view(parent)
 {
     setFlags(flags() | ItemIsFocusScope);
     m_slideAnim = new QPropertyAnimation(this);
     m_slideAnim->setTargetObject(this);
     m_slideAnim->setPropertyName("x");
-    //NOTE: the duration will be taked from kirigami units upon classBegin
+    // NOTE: the duration will be taken from kirigami units upon classBegin
     m_slideAnim->setDuration(0);
     m_slideAnim->setEasingCurve(QEasingCurve(QEasingCurve::InOutQuad));
-    connect(m_slideAnim, &QPropertyAnimation::finished, this, [this] () {
+    connect(m_slideAnim, &QPropertyAnimation::finished, this, [this]() {
         if (!m_view->currentItem()) {
             m_view->setCurrentIndex(m_items.indexOf(m_viewAnchorItem));
         } else {
@@ -309,7 +324,8 @@ ContentItem::ContentItem(ColumnView *parent)
 }
 
 ContentItem::~ContentItem()
-{}
+{
+}
 
 void ContentItem::setBoundedX(qreal x)
 {
@@ -317,7 +333,7 @@ void ContentItem::setBoundedX(qreal x)
         return;
     }
     m_slideAnim->stop();
-    setX(qRound(qBound(qMin(0.0, -width()+parentItem()->width()), x, 0.0)));
+    setX(qRound(qBound(qMin(0.0, -width() + parentItem()->width()), x, 0.0)));
 }
 
 void ContentItem::animateX(qreal newX)
@@ -326,7 +342,7 @@ void ContentItem::animateX(qreal newX)
         return;
     }
 
-    const qreal to = qRound(qBound(qMin(0.0, -width()+parentItem()->width()), newX, 0.0));
+    const qreal to = qRound(qBound(qMin(0.0, -width() + parentItem()->width()), newX, 0.0));
 
     m_slideAnim->stop();
     m_slideAnim->setStartValue(x());
@@ -342,20 +358,22 @@ void ContentItem::snapToItem()
     }
     QQuickItem *nextItem = childAt(firstItem->x() + firstItem->width() + 1, 0);
 
-    //need to make the last item visible?
-    if (nextItem && ((m_view->dragging() && m_lastDragDelta < 0) ||
-        (!m_view->dragging() && width() - (viewportRight()) < viewportLeft() - firstItem->x()))) {
+    // need to make the last item visible?
+    if (nextItem && //
+        ((m_view->dragging() && m_lastDragDelta < 0) //
+         || (!m_view->dragging() //
+             && (width() - viewportRight()) < (viewportLeft() - firstItem->x())))) {
         m_viewAnchorItem = nextItem;
         animateX(-nextItem->x() + m_leftPinnedSpace);
 
-    //The first one found?
-    } else if ((m_view->dragging() && m_lastDragDelta >= 0) ||
-        (!m_view->dragging() && viewportLeft() <= firstItem->x() + firstItem->width()/2) ||
-        !nextItem) {
+        // The first one found?
+    } else if ((m_view->dragging() && m_lastDragDelta >= 0) //
+               || (!m_view->dragging() && (viewportLeft() <= (firstItem->x() + (firstItem->width() / 2)))) //
+               || !nextItem) {
         m_viewAnchorItem = firstItem;
         animateX(-firstItem->x() + m_leftPinnedSpace);
 
-    //the second?
+        // the second?
     } else {
         m_viewAnchorItem = nextItem;
         animateX(-nextItem->x() + m_leftPinnedSpace);
@@ -389,9 +407,9 @@ qreal ContentItem::childWidth(QQuickItem *child)
     } else if (m_columnResizeMode == ColumnView::FixedColumns) {
         return qRound(qMin(parentItem()->width(), m_columnWidth));
 
-    // DynamicColumns
+        // DynamicColumns
     } else {
-        //TODO:look for Layout size hints
+        // TODO:look for Layout size hints
         qreal width = child->implicitWidth();
 
         if (width < 1.0) {
@@ -413,9 +431,17 @@ void ContentItem::layoutItems()
     int i = 0;
     m_leftPinnedSpace = 0;
     m_rightPinnedSpace = 0;
-    for (QQuickItem *child : m_items) {
+
+    bool reverse = qApp->layoutDirection() == Qt::RightToLeft;
+    auto it = !reverse ? m_items.begin() : m_items.end();
+    int increment = reverse ? -1 : +1;
+    auto lastPos = reverse ? m_items.begin() : m_items.end();
+
+    for (; it != lastPos; it += increment) {
+        // for (QQuickItem *child : std::as_const(m_items)) {
+        QQuickItem *child = reverse ? *(it - 1) : *it;
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(child, true));
-    
+
         if (child->isVisible()) {
             if (attached->isPinned() && m_view->columnResizeMode() != ColumnView::SingleColumn) {
                 QQuickItem *sep = nullptr;
@@ -453,7 +479,11 @@ void ContentItem::layoutItems()
             }
         }
 
-        attached->setIndex(i++);
+        if (reverse) {
+            attached->setIndex(m_items.count() - (++i));
+        } else {
+            attached->setIndex(i++);
+        }
 
         implicitWidth += child->implicitWidth();
 
@@ -488,9 +518,9 @@ void ContentItem::layoutPinnedItems()
     m_leftPinnedSpace = 0;
     m_rightPinnedSpace = 0;
 
-    for (QQuickItem *child : m_items) {
+    for (QQuickItem *child : std::as_const(m_items)) {
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(child, true));
-    
+
         if (child->isVisible()) {
             if (attached->isPinned()) {
                 QQuickItem *sep = nullptr;
@@ -516,18 +546,23 @@ void ContentItem::layoutPinnedItems()
 
 void ContentItem::updateVisibleItems()
 {
-    QList <QObject *> newItems;
+    QList<QObject *> newItems;
 
-    for (auto *item : m_items) {
-        if (item->isVisible() && item->x() + x() < width() && item->x() + item->width() + x() > 0) {
+    for (auto *item : std::as_const(m_items)) {
+        ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
+
+        if (item->isVisible() && item->x() + x() < m_view->width() && item->x() + item->width() + x() > 0) {
             newItems << item;
             connect(item, &QObject::destroyed, this, [this, item] {
                 m_visibleItems.removeAll(item);
             });
+            attached->setInViewport(true);
+        } else {
+            attached->setInViewport(false);
         }
     }
 
-    for (auto *item : m_visibleItems) {
+    for (auto *item : std::as_const(m_visibleItems)) {
         disconnect(item, &QObject::destroyed, this, nullptr);
     }
     const QQuickItem *oldFirstVisibleItem = m_visibleItems.isEmpty() ? nullptr : qobject_cast<QQuickItem *>(m_visibleItems.first());
@@ -535,12 +570,12 @@ void ContentItem::updateVisibleItems()
 
     if (newItems != m_visibleItems) {
         m_visibleItems = newItems;
-        emit m_view->visibleItemsChanged();
+        Q_EMIT m_view->visibleItemsChanged();
         if (!newItems.isEmpty() && m_visibleItems.first() != oldFirstVisibleItem) {
-            emit m_view->firstVisibleItemChanged();
+            Q_EMIT m_view->firstVisibleItemChanged();
         }
         if (!newItems.isEmpty() && m_visibleItems.last() != oldLastVisibleItem) {
-            emit m_view->lastVisibleItemChanged();
+            Q_EMIT m_view->lastVisibleItemChanged();
         }
     }
 }
@@ -574,20 +609,20 @@ void ContentItem::forgetItem(QQuickItem *item)
     updateVisibleItems();
     m_shouldAnimate = true;
     m_view->polish();
-    item->setVisible(false);
 
     if (index <= m_view->currentIndex()) {
         m_view->setCurrentIndex(qBound(0, index - 1, m_items.count() - 1));
     }
-    emit m_view->countChanged();
+    Q_EMIT m_view->countChanged();
 }
 
 QQuickItem *ContentItem::ensureSeparator(QQuickItem *item)
 {
     QQuickItem *separatorItem = m_separators.value(item);
-    
+
     if (!separatorItem) {
-        separatorItem = qobject_cast<QQuickItem *>(privateQmlComponentsPoolSelf->instance(qmlEngine(item))->m_separatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
+        separatorItem = qobject_cast<QQuickItem *>(
+            privateQmlComponentsPoolSelf->instance(qmlEngine(item))->m_separatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
         if (separatorItem) {
             separatorItem->setParentItem(item);
             separatorItem->setZ(9999);
@@ -603,9 +638,10 @@ QQuickItem *ContentItem::ensureSeparator(QQuickItem *item)
 QQuickItem *ContentItem::ensureRightSeparator(QQuickItem *item)
 {
     QQuickItem *separatorItem = m_rightSeparators.value(item);
-    
+
     if (!separatorItem) {
-        separatorItem = qobject_cast<QQuickItem *>(QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_rightSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
+        separatorItem = qobject_cast<QQuickItem *>(
+            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_rightSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
         if (separatorItem) {
             separatorItem->setParentItem(item);
             separatorItem->setZ(9999);
@@ -625,10 +661,9 @@ void ContentItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::It
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(value.item, true));
         attached->setView(m_view);
 
-        //connect(attached, &ColumnViewAttached::fillWidthChanged, m_view, &ColumnView::polish);
-         connect(attached, &ColumnViewAttached::fillWidthChanged, this, [this, attached](){
-             m_view->polish();
-             
+        // connect(attached, &ColumnViewAttached::fillWidthChanged, m_view, &ColumnView::polish);
+        connect(attached, &ColumnViewAttached::fillWidthChanged, this, [this] {
+            m_view->polish();
         });
         connect(attached, &ColumnViewAttached::reservedSpaceChanged, m_view, &ColumnView::polish);
 
@@ -649,7 +684,7 @@ void ContentItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::It
 
         m_shouldAnimate = true;
         m_view->polish();
-        emit m_view->countChanged();
+        Q_EMIT m_view->countChanged();
         break;
     }
     case QQuickItem::ItemChildRemovedChange: {
@@ -681,7 +716,7 @@ void ContentItem::syncItemsOrder()
     }
 
     m_items = childItems();
-    //NOTE: polish() here sometimes gets indefinitely delayed and items chaging order isn't seen
+    // NOTE: polish() here sometimes gets indefinitely delayed and items changing order isn't seen
     layoutItems();
 }
 
@@ -714,24 +749,19 @@ void ContentItem::updateRepeaterModel()
     }
 }
 
-
-
-
-
-
-
 ColumnView::ColumnView(QQuickItem *parent)
-    : QQuickItem(parent),
-      m_contentItem(nullptr)
+    : QQuickItem(parent)
+    , m_contentItem(nullptr)
 {
-    //NOTE: this is to *not* trigger itemChange
+    // NOTE: this is to *not* trigger itemChange
     m_contentItem = new ContentItem(this);
     setAcceptedMouseButtons(Qt::LeftButton | Qt::BackButton | Qt::ForwardButton);
+    setAcceptTouchEvents(false); // Relies on synthetized mouse events
     setFiltersChildMouseEvents(true);
 
-    connect(m_contentItem->m_slideAnim, &QPropertyAnimation::finished, this, [this] () {
+    connect(m_contentItem->m_slideAnim, &QPropertyAnimation::finished, this, [this]() {
         m_moving = false;
-        emit movingChanged();
+        Q_EMIT movingChanged();
     });
     connect(m_contentItem, &ContentItem::widthChanged, this, &ColumnView::contentWidthChanged);
     connect(m_contentItem, &ContentItem::xChanged, this, &ColumnView::contentXChanged);
@@ -768,7 +798,7 @@ void ColumnView::setColumnResizeMode(ColumnResizeMode mode)
     }
     m_contentItem->m_shouldAnimate = false;
     polish();
-    emit columnResizeModeChanged();
+    Q_EMIT columnResizeModeChanged();
 }
 
 qreal ColumnView::columnWidth() const
@@ -788,7 +818,7 @@ void ColumnView::setColumnWidth(qreal width)
     m_contentItem->m_columnWidth = width;
     m_contentItem->m_shouldAnimate = false;
     polish();
-    emit columnWidthChanged();
+    Q_EMIT columnWidthChanged();
 }
 
 int ColumnView::currentIndex() const
@@ -813,23 +843,19 @@ void ColumnView::setCurrentIndex(int index)
         m_currentItem->forceActiveFocus();
 
         // If the current item is not on view, scroll
-        QRectF mappedCurrent = m_currentItem->mapRectToItem(this,
-                                            QRectF(QPointF(0, 0),
-                                                   m_currentItem->size()));
+        QRectF mappedCurrent = m_currentItem->mapRectToItem(this, QRectF(QPointF(0, 0), m_currentItem->size()));
 
         if (m_contentItem->m_slideAnim->state() == QAbstractAnimation::Running) {
             mappedCurrent.moveLeft(mappedCurrent.left() + m_contentItem->x() + m_contentItem->m_slideAnim->endValue().toInt());
         }
 
-        //m_contentItem->m_slideAnim->stop();
+        // m_contentItem->m_slideAnim->stop();
 
-        QRectF contentsRect(m_contentItem->m_leftPinnedSpace,
+        QRectF contentsRect(m_contentItem->m_leftPinnedSpace, //
                             0,
                             width() - m_contentItem->m_rightPinnedSpace - m_contentItem->m_leftPinnedSpace,
                             height());
 
-        m_contentItem->m_shouldAnimate = true;
-
         if (!m_mouseDown) {
             if (!contentsRect.contains(mappedCurrent)) {
                 m_contentItem->m_viewAnchorItem = m_currentItem;
@@ -840,8 +866,8 @@ void ColumnView::setCurrentIndex(int index)
         }
     }
 
-    emit currentIndexChanged();
-    emit currentItemChanged();
+    Q_EMIT currentIndexChanged();
+    Q_EMIT currentItemChanged();
 }
 
 QQuickItem *ColumnView::currentItem()
@@ -849,7 +875,7 @@ QQuickItem *ColumnView::currentItem()
     return m_currentItem;
 }
 
-QList<QObject *>ColumnView::visibleItems() const
+QList<QObject *> ColumnView::visibleItems() const
 {
     return m_contentItem->m_visibleItems;
 }
@@ -890,7 +916,7 @@ void ColumnView::setTopPadding(qreal padding)
 
     m_topPadding = padding;
     polish();
-    emit topPaddingChanged();
+    Q_EMIT topPaddingChanged();
 }
 
 qreal ColumnView::bottomPadding() const
@@ -906,10 +932,9 @@ void ColumnView::setBottomPadding(qreal padding)
 
     m_bottomPadding = padding;
     polish();
-    emit bottomPaddingChanged();
+    Q_EMIT bottomPaddingChanged();
 }
 
-
 QQuickItem *ColumnView::contentItem() const
 {
     return m_contentItem;
@@ -929,7 +954,7 @@ void ColumnView::setScrollDuration(int duration)
     }
 
     m_contentItem->m_slideAnim->setDuration(duration);
-    emit scrollDurationChanged();
+    Q_EMIT scrollDurationChanged();
 }
 
 bool ColumnView::separatorVisible() const
@@ -946,7 +971,7 @@ void ColumnView::setSeparatorVisible(bool visible)
     m_separatorVisible = visible;
 
     if (visible) {
-        for (QQuickItem *item : m_contentItem->m_items) {
+        for (QQuickItem *item : std::as_const(m_contentItem->m_items)) {
             QQuickItem *sep = m_contentItem->ensureSeparator(item);
             if (sep) {
                 sep->setVisible(true);
@@ -962,15 +987,15 @@ void ColumnView::setSeparatorVisible(bool visible)
         }
 
     } else {
-        for (QQuickItem *sep : m_contentItem->m_separators.values()) {
+        for (QQuickItem *sep : std::as_const(m_contentItem->m_separators)) {
             sep->setVisible(false);
         }
-        for (QQuickItem *sep : m_contentItem->m_rightSeparators.values()) {
+        for (QQuickItem *sep : std::as_const(m_contentItem->m_rightSeparators)) {
             sep->setVisible(false);
         }
     }
 
-    emit separatorVisibleChanged();
+    Q_EMIT separatorVisibleChanged();
 }
 
 bool ColumnView::dragging() const
@@ -1014,14 +1039,40 @@ void ColumnView::setInteractive(bool interactive)
     if (!m_interactive) {
         if (m_dragging) {
             m_dragging = false;
-            emit draggingChanged();
+            Q_EMIT draggingChanged();
         }
 
         m_contentItem->snapToItem();
         setKeepMouseGrab(false);
     }
 
-    emit interactiveChanged();
+    Q_EMIT interactiveChanged();
+}
+
+bool ColumnView::acceptsMouse() const
+{
+    return m_acceptsMouse;
+}
+
+void ColumnView::setAcceptsMouse(bool accepts)
+{
+    if (m_acceptsMouse == accepts) {
+        return;
+    }
+
+    m_acceptsMouse = accepts;
+
+    if (!m_acceptsMouse) {
+        if (m_dragging) {
+            m_dragging = false;
+            Q_EMIT draggingChanged();
+        }
+
+        m_contentItem->snapToItem();
+        setKeepMouseGrab(false);
+    }
+
+    Q_EMIT acceptsMouseChanged();
 }
 
 void ColumnView::addItem(QQuickItem *item)
@@ -1046,25 +1097,83 @@ void ColumnView::insertItem(int pos, QQuickItem *item)
     item->setParentItem(m_contentItem);
 
     item->forceActiveFocus();
-    // We layout immediately to be sure all geometries are final after the return of this call
-    m_contentItem->m_shouldAnimate = false;
+
+    // Animate shift to new item.
+    m_contentItem->m_shouldAnimate = true;
     m_contentItem->layoutItems();
-    emit contentChildrenChanged();
+    Q_EMIT contentChildrenChanged();
 
     // In order to keep the same current item we need to increase the current index if displaced
     // NOTE: just updating m_currentIndex does *not* update currentItem (which is what we need atm) while setCurrentIndex will update also currentItem
     if (m_currentIndex >= pos) {
         ++m_currentIndex;
-        emit currentIndexChanged();
+        Q_EMIT currentIndexChanged();
     }
 
-    emit itemInserted(pos, item);
+    Q_EMIT itemInserted(pos, item);
+}
+
+void ColumnView::replaceItem(int pos, QQuickItem *item)
+{
+    if (pos < 0 || pos >= m_contentItem->m_items.length()) {
+        qCWarning(KirigamiLog) << "Position" << pos << "passed to ColumnView::replaceItem is out of range.";
+        return;
+    }
+
+    if (!item) {
+        qCWarning(KirigamiLog) << "Null item passed to ColumnView::replaceItem.";
+        return;
+    }
+
+    QQuickItem *oldItem = m_contentItem->m_items[pos];
+
+    // In order to keep the same current item we need to increase the current index if displaced
+    if (m_currentIndex >= pos) {
+        setCurrentIndex(m_currentIndex - 1);
+    }
+
+    m_contentItem->forgetItem(oldItem);
+    oldItem->setVisible(false);
+
+    ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(oldItem, false));
+
+    if (attached && attached->shouldDeleteOnRemove()) {
+        oldItem->deleteLater();
+    } else {
+        oldItem->setParentItem(attached ? attached->originalParent() : nullptr);
+    }
+
+    Q_EMIT itemRemoved(oldItem);
+
+    if (!m_contentItem->m_items.contains(item)) {
+        m_contentItem->m_items.insert(qBound(0, pos, m_contentItem->m_items.length()), item);
+
+        connect(item, &QObject::destroyed, m_contentItem, [this, item]() {
+            removeItem(item);
+        });
+        ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
+        attached->setOriginalParent(item->parentItem());
+        attached->setShouldDeleteOnRemove(item->parentItem() == nullptr && QQmlEngine::objectOwnership(item) == QQmlEngine::JavaScriptOwnership);
+        item->setParentItem(m_contentItem);
+
+        if (m_currentIndex >= pos) {
+            ++m_currentIndex;
+            Q_EMIT currentIndexChanged();
+        }
+
+        Q_EMIT itemInserted(pos, item);
+    }
+
+    // Disable animation so replacement happens immediately.
+    m_contentItem->m_shouldAnimate = false;
+    m_contentItem->layoutItems();
+    Q_EMIT contentChildrenChanged();
 }
 
 void ColumnView::moveItem(int from, int to)
 {
-    if (m_contentItem->m_items.isEmpty()
-        || from < 0 || from >= m_contentItem->m_items.length()
+    if (m_contentItem->m_items.isEmpty() //
+        || from < 0 || from >= m_contentItem->m_items.length() //
         || to < 0 || to >= m_contentItem->m_items.length()) {
         return;
     }
@@ -1074,13 +1183,13 @@ void ColumnView::moveItem(int from, int to)
 
     if (from == m_currentIndex) {
         m_currentIndex = to;
-        emit currentIndexChanged();
+        Q_EMIT currentIndexChanged();
     } else if (from < m_currentIndex && to > m_currentIndex) {
         --m_currentIndex;
-        emit currentIndexChanged();
+        Q_EMIT currentIndexChanged();
     } else if (from > m_currentIndex && to <= m_currentIndex) {
         ++m_currentIndex;
-        emit currentIndexChanged();
+        Q_EMIT currentIndexChanged();
     }
 
     polish();
@@ -1111,6 +1220,7 @@ QQuickItem *ColumnView::removeItem(QQuickItem *item)
     }
 
     m_contentItem->forgetItem(item);
+    item->setVisible(false);
 
     ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, false));
 
@@ -1120,19 +1230,19 @@ QQuickItem *ColumnView::removeItem(QQuickItem *item)
         item->setParentItem(attached ? attached->originalParent() : nullptr);
     }
 
-    emit itemRemoved(item);
+    Q_EMIT itemRemoved(item);
 
     return item;
 }
 
 QQuickItem *ColumnView::removeItem(int pos)
 {
-    if (m_contentItem->m_items.isEmpty()
+    if (m_contentItem->m_items.isEmpty() //
         || pos < 0 || pos >= m_contentItem->m_items.length()) {
         return nullptr;
     }
 
-    return removeItem(m_contentItem->m_items[pos]); 
+    return removeItem(m_contentItem->m_items[pos]);
 }
 
 QQuickItem *ColumnView::pop(QQuickItem *item)
@@ -1151,11 +1261,11 @@ QQuickItem *ColumnView::pop(QQuickItem *item)
 
 void ColumnView::clear()
 {
-    for (QQuickItem *item : m_contentItem->m_items) {
+    for (QQuickItem *item : std::as_const(m_contentItem->m_items)) {
         removeItem(item);
     }
     m_contentItem->m_items.clear();
-    emit contentChildrenChanged();
+    Q_EMIT contentChildrenChanged();
 }
 
 bool ColumnView::containsItem(QQuickItem *item)
@@ -1198,18 +1308,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             return false;
         }
 
-        m_contentItem->m_slideAnim->stop();
-        if (item->property("preventStealing").toBool()) {
-            m_contentItem->snapToItem();
-            return false;
-        }
-        m_oldMouseX = m_startMouseX = mapFromItem(item, me->localPos()).x();
-
-        m_mouseDown = true;
-        me->setAccepted(false);
-        setKeepMouseGrab(false);
-
-        // On press, we set the current index of the view to the root item 
+        // On press, we set the current index of the view to the root item
         QQuickItem *candidateItem = item;
         while (candidateItem->parentItem() && candidateItem->parentItem() != m_contentItem) {
             candidateItem = candidateItem->parentItem();
@@ -1217,19 +1316,41 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
         if (candidateItem->parentItem() == m_contentItem) {
             setCurrentIndex(m_contentItem->m_items.indexOf(candidateItem));
         }
+
+        // if !m_acceptsMouse we don't drag with mouse
+        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
+            event->setAccepted(false);
+            return false;
+        }
+
+        m_contentItem->m_slideAnim->stop();
+        if (item->property("preventStealing").toBool()) {
+            m_contentItem->snapToItem();
+            return false;
+        }
+        m_oldMouseX = m_startMouseX = mapFromItem(item, me->localPos()).x();
+        m_oldMouseY = m_startMouseY = mapFromItem(item, me->localPos()).y();
+
+        m_mouseDown = true;
+        me->setAccepted(false);
+        setKeepMouseGrab(false);
+
         break;
     }
     case QEvent::MouseMove: {
         QMouseEvent *me = static_cast<QMouseEvent *>(event);
 
+        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
+            return false;
+        }
+
         if (!(me->buttons() & Qt::LeftButton)) {
             return false;
         }
 
-        if ((!keepMouseGrab() && item->keepMouseGrab()) || item->property("preventStealing").toBool()) {
-            m_contentItem->snapToItem();
-            return false;
-        }
+        const QPointF pos = mapFromItem(item, me->localPos());
+
+        bool verticalScrollIntercepted = false;
 
         QQuickItem *candidateItem = item;
         while (candidateItem->parentItem() && candidateItem->parentItem() != m_contentItem) {
@@ -1242,7 +1363,26 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             }
         }
 
-        const QPointF pos = mapFromItem(item, me->localPos());
+        {
+            ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(candidateItem, true));
+
+            ScrollIntentionEvent scrollIntentionEvent;
+            scrollIntentionEvent.delta = QPointF(pos.x() - m_oldMouseX, pos.y() - m_oldMouseY);
+
+            Q_EMIT attached->scrollIntention(&scrollIntentionEvent);
+
+            if (scrollIntentionEvent.accepted) {
+                verticalScrollIntercepted = true;
+                event->setAccepted(true);
+            }
+        }
+
+        if ((!keepMouseGrab() && item->keepMouseGrab()) || item->property("preventStealing").toBool()) {
+            m_contentItem->snapToItem();
+            m_oldMouseX = pos.x();
+            m_oldMouseY = pos.y();
+            return false;
+        }
 
         const bool wasDragging = m_dragging;
         // If a drag happened, start to steal all events, use startDragDistance * 2 to give time to widgets to take the mouse grab by themselves
@@ -1250,8 +1390,8 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
 
         if (m_dragging != wasDragging) {
             m_moving = true;
-            emit movingChanged();
-            emit draggingChanged();
+            Q_EMIT movingChanged();
+            Q_EMIT draggingChanged();
         }
 
         if (m_dragging) {
@@ -1260,15 +1400,21 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
 
         m_contentItem->m_lastDragDelta = pos.x() - m_oldMouseX;
         m_oldMouseX = pos.x();
+        m_oldMouseY = pos.y();
 
         setKeepMouseGrab(m_dragging);
         me->setAccepted(m_dragging);
-        return m_dragging;
+
+        return m_dragging && !verticalScrollIntercepted;
         break;
     }
     case QEvent::MouseButtonRelease: {
         QMouseEvent *me = static_cast<QMouseEvent *>(event);
 
+        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
+            return false;
+        }
+
         if (me->button() == Qt::BackButton && m_currentIndex > 0) {
             setCurrentIndex(m_currentIndex - 1);
             me->accept();
@@ -1289,7 +1435,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
         m_contentItem->m_lastDragDelta = 0;
         if (m_dragging) {
             m_dragging = false;
-            emit draggingChanged();
+            Q_EMIT draggingChanged();
         }
 
         if (item->property("preventStealing").toBool()) {
@@ -1298,7 +1444,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
 
         event->accept();
 
-        //if a drag happened, don't pass the event
+        // if a drag happened, don't pass the event
         const bool block = keepMouseGrab();
         setKeepMouseGrab(false);
 
@@ -1315,6 +1461,11 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
 
 void ColumnView::mousePressEvent(QMouseEvent *event)
 {
+    if (!m_acceptsMouse && event->source() == Qt::MouseEventNotSynthesized) {
+        event->setAccepted(false);
+        return;
+    }
+
     if (event->button() == Qt::BackButton || event->button() == Qt::ForwardButton) {
         event->accept();
         return;
@@ -1348,8 +1499,8 @@ void ColumnView::mouseMoveEvent(QMouseEvent *event)
     m_dragging = keepMouseGrab() || qAbs(event->localPos().x() - m_startMouseX) > qApp->styleHints()->startDragDistance() * 2;
     if (m_dragging != wasDragging) {
         m_moving = true;
-        emit movingChanged();
-        emit draggingChanged();
+        Q_EMIT movingChanged();
+        Q_EMIT draggingChanged();
     }
 
     setKeepMouseGrab(m_dragging);
@@ -1386,7 +1537,7 @@ void ColumnView::mouseReleaseEvent(QMouseEvent *event)
 
     if (m_dragging) {
         m_dragging = false;
-        emit draggingChanged();
+        Q_EMIT draggingChanged();
     }
 
     setKeepMouseGrab(false);
@@ -1404,7 +1555,7 @@ void ColumnView::mouseUngrabEvent()
 
     if (m_dragging) {
         m_dragging = false;
-        emit draggingChanged();
+        Q_EMIT draggingChanged();
     }
 
     setKeepMouseGrab(false);
@@ -1414,7 +1565,7 @@ void ColumnView::classBegin()
 {
     auto syncColumnWidth = [this]() {
         m_contentItem->m_columnWidth = privateQmlComponentsPoolSelf->instance(qmlEngine(this))->m_units->property("gridUnit").toInt() * 20;
-        emit columnWidthChanged();
+        Q_EMIT columnWidthChanged();
     };
 
     connect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::gridUnitChanged, this, syncColumnWidth);
@@ -1422,7 +1573,7 @@ void ColumnView::classBegin()
 
     auto syncDuration = [this]() {
         m_contentItem->m_slideAnim->setDuration(QmlComponentsPoolSingleton::instance(qmlEngine(this))->m_units->property("longDuration").toInt());
-        emit scrollDurationChanged();
+        Q_EMIT scrollDurationChanged();
     };
 
     connect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::longDurationChanged, this, syncDuration);
@@ -1511,11 +1662,12 @@ void ColumnView::contentChildren_clear(QQmlListProperty<QQuickItem> *prop)
 
 QQmlListProperty<QQuickItem> ColumnView::contentChildren()
 {
-    return QQmlListProperty<QQuickItem>(this, nullptr,
-                                     contentChildren_append,
-                                     contentChildren_count,
-                                     contentChildren_at,
-                                     contentChildren_clear);
+    return QQmlListProperty<QQuickItem>(this, //
+                                        nullptr,
+                                        contentChildren_append,
+                                        contentChildren_count,
+                                        contentChildren_at,
+                                        contentChildren_clear);
 }
 
 void ColumnView::contentData_append(QQmlListProperty<QObject> *prop, QObject *object)
@@ -1527,7 +1679,7 @@ void ColumnView::contentData_append(QQmlListProperty<QObject> *prop, QObject *ob
 
     view->m_contentData.append(object);
     QQuickItem *item = qobject_cast<QQuickItem *>(object);
-    //exclude repeaters from layout
+    // exclude repeaters from layout
     if (item && item->inherits("QQuickRepeater")) {
         item->setParentItem(view);
 
@@ -1585,7 +1737,8 @@ void ColumnView::contentData_clear(QQmlListProperty<QObject> *prop)
 
 QQmlListProperty<QObject> ColumnView::contentData()
 {
-    return QQmlListProperty<QObject>(this, nullptr,
+    return QQmlListProperty<QObject>(this, //
+                                     nullptr,
                                      contentData_append,
                                      contentData_count,
                                      contentData_at,
